package main

import (
	"bytes"
	"crypto/sha256"
	"fmt"
	"image"
	_ "image/gif"
	_ "image/jpeg"
	_ "image/png"
	"io"
	"log"
	"math/rand"
	"net/url"
	"os"
	"strings"
	"time"

	"github.com/mmcdole/gofeed"
)

func init() {
	rand.Seed(time.Now().Unix())
}

type Feed struct {
	Feedid        int64
	Feedurl       string
	Feedname      string
	Checkinterval time.Duration
	Nextcheck     time.Time
}

func getfeeds() []*Feed {
	rows, err := db.Query("select feedid, feedurl, feedname, checkinterval, nextcheck from feeds")
	if err != nil {
		log.Fatalf("error querying feeds: %s", err)
	}
	defer rows.Close()
	var feeds []*Feed
	for rows.Next() {
		var feed Feed
		var dt string
		var dur string
		err = rows.Scan(&feed.Feedid, &feed.Feedurl, &feed.Feedname, &dur, &dt)
		if err != nil {
			log.Printf("error scanning feed row: %s", err)
			continue
		}
		feed.Checkinterval, _ = time.ParseDuration(dur)
		feed.Nextcheck, _ = time.Parse(dbtimeformat, dt)
		feeds = append(feeds, &feed)
	}
	return feeds
}

func reverseItems(items []*gofeed.Item) {
	for i, j := 0, len(items)-1; i < j; i, j = i+1, j-1 {
		items[i], items[j] = items[j], items[i]
	}
}

func checkimg(url string) (string, string, int, int) {
	hasher := sha256.New()
	hasher.Write([]byte(url))
	var sum []byte
	sum = hasher.Sum(sum)
	name := fmt.Sprintf("%x", sum)

	rows, err := db.Query("select format, width, height from images where name = ?", name)
	if err != nil {
		log.Printf("unable to check for image: %s", err)
		return "", "", 0, 0
	}
	defer rows.Close()
	if rows.Next() {
		var format string
		var width int
		var height int
		err = rows.Scan(&format, &width, &height)
		if err != nil {
			log.Printf("error scanning img row: %s", err)
			return "", "", 0, 0
		}
		return name, format, width, height
	}
	return name, "", 0, 0
}

func saveimg(imgurl *url.URL) (string, string, int, int) {
	url := imgurl.String()
	log.Printf("collecting img %s", url)

	name, format, width, height := checkimg(url)
	if format != "" {
		return name, format, width, height
	}
	if name == "" {
		log.Printf("skipping img %s", url)
		return "", "", 0, 0
	}

	delay, _ := time.ParseDuration("1s")
	time.Sleep(delay)

	resp, err := httpget(url)
	if err != nil {
		log.Printf("error getting img %s: %s", url, err)
		return "", "", 0, 0
	}
	defer resp.Body.Close()
	var buf bytes.Buffer
	io.Copy(&buf, resp.Body)
	reader := bytes.NewReader(buf.Bytes())
	conf, format, err := image.DecodeConfig(reader)
	if err != nil {
		log.Printf("unknown image %s: %s", url, err)
		return "", "", 0, 0
	}
	_, err = db.Exec("insert into images (name, origurl, format, width, height, data) values (?, ?, ?, ?, ?, ?)", name, url, format, conf.Width, conf.Height, buf.Bytes())
	if err != nil {
		log.Printf("error saving img: %s", err)
		return "", "", 0, 0
	}

	return name, format, conf.Width, conf.Height
}

func havelink(feedid int64, link string) bool {
	if link == "" {
		log.Println("not saving empty link")
		return true
	}
	rows, err := db.Query("select linkid from links where feedid = ? and link = ?",
		feedid, link)
	if err != nil {
		log.Printf("unable to check for link: %s", err)
		return false
	}
	defer rows.Close()
	if rows.Next() {
		return true
	}
	return false
}

func savelink(feedid int64, link string, title string, dt *time.Time, content string) {
	res, err := db.Exec("insert into links (feedid, link, title, dt, content) values (?, ?, ?, ?, ?)",
		feedid, link, title, dt.UTC().Format(dbtimeformat), content)
	if err != nil {
		log.Printf("unable to save link: %s", err)
		return
	}
	linkid, _ := res.LastInsertId()
	_, err = db.Exec("insert into userlinks (userid, linkid, feedid, seen) values ((select userid from users), ?, ?, ?)",
		linkid, feedid, 0)
	if err != nil {
		log.Printf("unable to save userlinks: %s", err)
		return
	}
	log.Printf("saved new link %s", link)
}

func resolvelink(baseurl *url.URL, rel string) string {
	relurl, err := url.Parse(rel)
	if err != nil {
		return ""
	}
	absurl := baseurl.ResolveReference(relurl)
	return absurl.String()
}

func calcNextcheck(start time.Time, interval time.Duration) time.Time {
	drift := int64(interval.Seconds() / 3)
	drift = rand.Int63n(drift)

	return start.Add(interval + time.Duration(drift)*time.Second).UTC()
}

func checkfeeds() time.Duration {
	now := time.Now()
	waketime := now.Add(1 * time.Hour)
	feeds := getfeeds()
	for _, feed := range feeds {
		baseurl, err := url.Parse(feed.Feedurl)
		if err != nil {
			log.Printf("unable to parse feedurl %s: %s", feed.Feedurl, err)
			continue
		}
		if feed.Checkinterval == 0 {
			log.Printf("disabled feed: %s", feed.Feedurl)
			continue
		}
		if feed.Nextcheck.After(now) {
			if feed.Nextcheck.Before(waketime) {
				waketime = feed.Nextcheck
			}
			continue
		}
		feed.Nextcheck = calcNextcheck(now, feed.Checkinterval)
		_, err = db.Exec("update feeds set nextcheck = ? where feedid = ?",
			feed.Nextcheck.Format(dbtimeformat), feed.Feedid)
		if err != nil {
			log.Printf("unable to update checktime: %s", err)
			continue
		}
		if strings.HasPrefix(feed.Feedurl, "https://twitter.com/") {
			log.Println("grabbing tweets from", feed.Feedurl)
			loadtweets(feed, baseurl)
			delay, _ := time.ParseDuration("3s")
			time.Sleep(delay)
			continue
		}

		log.Println("feeding from", feed.Feedurl)
		var reader io.ReadCloser
		if baseurl.Scheme == "" {
			fd, err0 := os.Open(feed.Feedurl)
			err = err0
			if err == nil {
				reader = fd
			}
		} else {
			resp, err0 := httpget(feed.Feedurl)
			err = err0
			if err == nil {
				reader = resp.Body
			}
		}
		if err != nil {
			log.Println("error getting feed", err)
			continue
		}
		parser := gofeed.NewParser()
		rss, err := parser.Parse(reader)
		reader.Close()
		if err != nil {
			log.Println("error parsing feed", err)
			continue
		}
		reverseItems(rss.Items)
		for _, item := range rss.Items {
			link := item.Link
			if link == "" ||
				strings.Contains(link, "feedproxy.google.com") ||
				strings.Contains(link, "tracking.feedpress.it") {
				link = item.GUID
			}
			if link != "" {
				link = resolvelink(baseurl, link)
			}
			if havelink(feed.Feedid, link) {
				continue
			}
			title := item.Title
			dt := &now
			if item.UpdatedParsed != nil {
				dt = item.UpdatedParsed
			} else if item.PublishedParsed != nil {
				dt = item.PublishedParsed
			}
			content := item.Content
			if content == "" {
				content = item.Description
			}
			content = string(cleanstring(content, baseurl))
			savelink(feed.Feedid, link, title, dt, content)
		}
	}
	return waketime.Sub(now).Round(time.Second) + 5*time.Second
}

func feeder() {
	sleeper := time.NewTimer(0)
	for {
		select {
		case <-wakefeeder:
			if !sleeper.Stop() {
				<-sleeper.C
			}
		case <-sleeper.C:
		}
		sleeptime := checkfeeds()
		sleeper.Reset(sleeptime)
		log.Printf("next check: %s", sleeptime)
	}
}
