package main

import (
	"fmt"
	"html/template"
	"io"
	"log"
	"net/url"
	"sort"
	"strings"

	"golang.org/x/net/html"
)

var permittedtags = []string{"div", "h1", "h2", "h3", "h4", "h5", "h6",
	"table", "thead", "tbody", "th", "tr", "td",
	"p", "br", "pre", "code", "blockquote", "strong", "em", "b", "i", "s", "sup",
	"ol", "ul", "li"}
var permittedattr = []string{"colspan", "rowspan"}
var bannedtags = []string{"script", "style"}
var bannedimghosts = []string{"feeds.feedburner.com"}

func init() {
	sort.Strings(permittedtags)
	sort.Strings(permittedattr)
	sort.Strings(bannedtags)
	sort.Strings(bannedimghosts)
}

func contains(array []string, tag string) bool {
	idx := sort.SearchStrings(array, tag)
	return idx < len(array) && array[idx] == tag
}

func getattr(node *html.Node, attr string) string {
	for _, a := range node.Attr {
		if a.Key == attr {
			return a.Val
		}
	}
	return ""
}

func hasclass(node *html.Node, class string) bool {
	return strings.Contains(" "+getattr(node, "class")+" ", " "+class+" ")
}

func writetag(w io.Writer, node *html.Node) {
	io.WriteString(w, "<")
	io.WriteString(w, node.Data)
	for _, attr := range node.Attr {
		if contains(permittedattr, attr.Key) {
			io.WriteString(w, fmt.Sprintf(` %s="%s"`, attr.Key,
				html.EscapeString(attr.Val)))
		}
	}
	io.WriteString(w, ">")
}

func render(w io.Writer, node *html.Node, baseurl *url.URL) {
	switch node.Type {
	case html.ElementNode:
		tag := node.Data
		switch {
		case tag == "a":
			if getattr(node, "data-pre-embedded") != "" {
				return
			}
			href := getattr(node, "href")
			if data_url := getattr(node, "data-expanded-url"); data_url != "" {
				if strings.HasPrefix(data_url, "https://twitter.com/") &&
					hasclass(node, "u-hidden") {
					return
				}
				href = data_url
			} else {
				hrefurl, err := url.Parse(href)
				if err != nil {
					href = "#BROKEN-" + href
				} else {
					hrefurl = baseurl.ResolveReference(hrefurl)
					href = hrefurl.String()
				}
			}
			io.WriteString(w, fmt.Sprintf(`<a href="%s" rel=noreferrer>`,
				html.EscapeString(href)))
		case tag == "img":
			div := replaceimg(baseurl, node)
			if div != "skip" {
				io.WriteString(w, div)
			}
		case tag == "span":
			if hasclass(node, "tco-ellipsis") {
				return
			}
			if hasclass(node, "Icon--verified") {
				return
			}
		case tag == "iframe":
			io.WriteString(w, html.EscapeString(
				fmt.Sprintf(`<iframe src="%s">`, getattr(node, "src"))))
		case contains(permittedtags, tag):
			writetag(w, node)
		case contains(bannedtags, tag):
			return
		}
	case html.TextNode:
		io.WriteString(w, html.EscapeString(node.Data))
	}
	for c := node.FirstChild; c != nil; c = c.NextSibling {
		render(w, c, baseurl)
	}
	if node.Type == html.ElementNode {
		tag := node.Data
		if tag == "a" || (contains(permittedtags, tag) && tag != "br") {
			io.WriteString(w, fmt.Sprintf("</%s>", tag))
		}
		if tag == "p" || tag == "div" {
			io.WriteString(w, "\n")
		}
	}
}

func replaceimg(baseurl *url.URL, node *html.Node) string {
	src := getattr(node, "src")
	alt := getattr(node, "alt")
	title := getattr(node, "title")
	if hasclass(node, "Emoji") && alt != "" {
		return html.EscapeString(alt)
	}
	return createimgloader(baseurl, src, alt, title)
}

func createimgloader(baseurl *url.URL, src string, alt string, title string) string {
	if strings.HasPrefix(src, "data:") {
		return fmt.Sprintf(`<img src="%s" alt="%s" title="%s">`,
			html.EscapeString(src), html.EscapeString(alt),
			html.EscapeString(title))
	}
	srcurl, err := url.Parse(src)
	if err != nil {
		return fmt.Sprintf(" (image %s failed to parse) ",
			html.EscapeString(src))
	}
	srcurl = baseurl.ResolveReference(srcurl)
	if contains(bannedimghosts, srcurl.Host) {
		return "skip"
	}
	name, format, w, h := saveimg(srcurl)
	if name == "" {
		return fmt.Sprintf(" (image %s failed to load) ",
			html.EscapeString(src))
	}
	if w > 960 {
		r := 960 / float64(w)
		w = int(float64(w) * r)
		h = int(float64(h) * r)
	}

	img := fmt.Sprintf(`<img src="/images/%s.%s" alt="%s" title="%s" width=%d height=%d>`,
		name, format, html.EscapeString(alt), html.EscapeString(title), w, h)
	div := fmt.Sprintf(`<div onclick="loadimg(this); return false;"><div class="imgholder" style="width: %dpx; height: %dpx" data="%s">[ load image %s ]</div></div>`, w, h, html.EscapeString(img), html.EscapeString(srcurl.String()))
	return div
}

func cleannode(node *html.Node, baseurl *url.URL) template.HTML {
	var buf strings.Builder
	render(&buf, node, baseurl)
	return template.HTML(buf.String())
}

func cleanstring(shtml string, baseurl *url.URL) template.HTML {
	reader := strings.NewReader(shtml)
	body, err := html.Parse(reader)
	if err != nil {
		log.Printf("error parsing %s: %s", baseurl.String(), err)
		return ""
	}
	return cleannode(body, baseurl)
}
