package main

import (
	"context"
	"crypto/rand"
	"crypto/sha512"
	"fmt"
	"log"
	"net/http"
	"regexp"

	"golang.org/x/crypto/bcrypt"
)

type keytype struct{}

var thekey keytype

type UserInfo struct {
	Username   string
	UserID     int64
	LastUnread int64
}

func LoginWrap(handler http.HandlerFunc) http.HandlerFunc {
	return func(w http.ResponseWriter, r *http.Request) {
		userinfo, ok := checkauth(r)
		if !ok {
			loginredirect(w, r)
			return
		}
		ctx := context.WithValue(r.Context(), thekey, userinfo)
		handler(w, r.WithContext(ctx))
	}
}

func GetUserInfo(r *http.Request) *UserInfo {
	userinfo, ok := r.Context().Value(thekey).(*UserInfo)
	if !ok {
		return nil
	}
	return userinfo
}

func loginredirect(w http.ResponseWriter, r *http.Request) {
	http.Redirect(w, r, "/login", http.StatusSeeOther)
}

var authregex = regexp.MustCompile("^[[:alnum:]]+$")
var authlen = 64

func checkauth(r *http.Request) (*UserInfo, bool) {
	cookie, err := r.Cookie("auth")
	if err != nil {
		log.Printf("no cookie")
		return nil, false
	}
	auth := cookie.Value
	if !(len(auth) == authlen && authregex.MatchString(auth)) {
		log.Printf("bad auth")
		return nil, false
	}
	hasher := sha512.New512_256()
	hasher.Write([]byte(auth))
	var sum []byte
	sum = hasher.Sum(sum)
	hashauth := fmt.Sprintf("%x", sum)
	rows, err := db.Query("select userid, username, lastunread from users where userid = (select userid from auth where hash = ?)", hashauth)
	if err != nil {
		log.Printf("unable to select auth: %s", err)
		return nil, false
	}
	defer rows.Close()
	if !rows.Next() {
		log.Printf("no auth found")
		return nil, false
	}
	userinfo := &UserInfo{}
	err = rows.Scan(&userinfo.UserID, &userinfo.Username, &userinfo.LastUnread)
	if err != nil {
		log.Printf("error scanning auth row: %s", err)
		return nil, false
	}
	return userinfo, true
}

func loaduser(username string) (int64, string, bool) {
	rows, err := db.Query("select userid, hash from users where username = ?", username)
	if err != nil {
		log.Printf("error querying username: %s", err)
		return -1, "", false
	}
	defer rows.Close()
	if !rows.Next() {
		log.Printf("no username found")
		return -1, "", false
	}
	var userid int64
	var hash string
	err = rows.Scan(&userid, &hash)
	if err != nil {
		log.Printf("error loading username: %s", err)
		return -1, "", false
	}
	return userid, hash, true

}

var userregex = regexp.MustCompile("^[[:alnum:]]+$")
var userlen = 32
var passlen = 128

func dologin(w http.ResponseWriter, r *http.Request) {
	username := r.FormValue("username")
	password := r.FormValue("password")

	if len(username) == 0 || len(username) > userlen ||
		!userregex.MatchString(username) || len(password) == 0 ||
		len(password) > passlen {
		log.Printf("invalid password attempt")
		loginredirect(w, r)
		return
	}
	userid, hash, ok := loaduser(username)
	if !ok {
		loginredirect(w, r)
		return
	}

	err := bcrypt.CompareHashAndPassword([]byte(hash), []byte(password))
	if err != nil {
		log.Printf("incorrect password")
		loginredirect(w, r)
		return
	}
	var randbytes [64]byte
	rand.Read(randbytes[:])
	hasher := sha512.New512_256()
	hasher.Write(randbytes[:])
	var sum []byte
	sum = hasher.Sum(sum)
	hash = fmt.Sprintf("%x", sum)

	http.SetCookie(w, &http.Cookie{
		Name:     "auth",
		Value:    hash,
		MaxAge:   3600 * 24 * 30,
		Secure:   false,
		HttpOnly: true,
	})

	hasher.Reset()
	hasher.Write([]byte(hash))
	sum = []byte{}
	sum = hasher.Sum(sum)
	hash = fmt.Sprintf("%x", sum)

	_, err = db.Exec("insert into auth (userid, hash) values (?, ?)", userid, hash)
	if err != nil {
		log.Printf("error saving auth: %s", err)
	}

	log.Printf("successful login")
	http.Redirect(w, r, "/", http.StatusSeeOther)
}

func dologout(w http.ResponseWriter, r *http.Request) {
	userinfo, ok := checkauth(r)
	if !ok {
		return
	}
	_, err := db.Exec("delete from auth where userid = ?", userinfo.UserID)
	if err != nil {
		log.Printf("error deleting old auth: %s", err)
		return
	}
	http.SetCookie(w, &http.Cookie{
		Name:     "auth",
		Value:    "x",
		MaxAge:   1,
		Secure:   false,
		HttpOnly: true,
	})
	http.Redirect(w, r, "/", http.StatusSeeOther)
}
