package main

import (
	"database/sql"
	"fmt"
	"html/template"
	"log"
	"net/http"
	"net/url"
	"regexp"
	"sort"
	"strconv"
	"strings"
	"time"

	"github.com/gorilla/mux"
)

type Link struct {
	Linkid   int64
	Feedname string
	Link     string
	Title    template.HTML
	Date     time.Time
	Content  template.HTML
}

var readviews = template.Must(template.ParseFiles("views/feeds.html",
	"views/links.html", "views/header.html"))

func feedspage(w http.ResponseWriter, r *http.Request) {
	feeds := getfeeds()

	sort.Slice(feeds, func(i, j int) bool {
		return feeds[i].Feedname < feeds[j].Feedname
	})

	err := readviews.ExecuteTemplate(w, "feeds.html", feeds)
	if err != nil {
		log.Printf("template error: %s", err)
	}
}

func getlinks(userinfo *UserInfo, feedid int64, lastlink int64, search string) ([]*Link, string) {
	query := "select links.linkid, feedname, link, title, dt, content from links join feeds on links.feedid = feeds.feedid where"
	var args []interface{}
	beforefmt := "/before/%d"
	if feedid > 0 {
		query += " links.feedid = ?"
		args = append(args, feedid)
		beforefmt = fmt.Sprintf("/feed/%d", feedid) + beforefmt
	} else if search != "" {
		if !regexp.MustCompile("^[[:alnum:] ]*$").MatchString(search) {
			search = ""
		}
		query += fmt.Sprintf(" links.content like '%%%s%%'", search)
		beforefmt = "/search/before/%d?q=" + url.QueryEscape(search)
	} else {
		query = query[:len(query)-5] + "join userlinks on links.linkid = userlinks.linkid where userlinks.userid = ? and seen = 0"
		args = append(args, userinfo.UserID)
		query += " and links.linkid >= ?"
		args = append(args, userinfo.LastUnread)
	}
	if lastlink > 0 {
		query += " and links.linkid < ?"
		args = append(args, lastlink)
	}

	query += " order by links.linkid desc limit 30"
	rows, err := db.Query(query, args...)
	if err != nil {
		log.Printf("error querying links: %s", err)
		return nil, ""
	}
	defer rows.Close()
	lastlink = 0
	var links []*Link
	for rows.Next() {
		var link Link
		var dt string
		var title string
		err = rows.Scan(&link.Linkid, &link.Feedname, &link.Link, &title, &dt, &link.Content)
		if err != nil {
			log.Printf("error loading link: %s", err)
			continue
		}
		link.Title = template.HTML(strings.Replace(title, "<", "&lt;", -1))
		link.Date, _ = time.Parse(dbtimeformat, dt)
		links = append(links, &link)
		lastlink = link.Linkid
	}
	beforelink := fmt.Sprintf(beforefmt, lastlink)

	return links, beforelink
}

func linkspage(w http.ResponseWriter, r *http.Request) {
	feedid, _ := strconv.ParseInt(mux.Vars(r)["feedid"], 10, 0)
	lastlink, _ := strconv.ParseInt(mux.Vars(r)["linkid"], 10, 0)
	search := r.FormValue("q")

	userinfo := GetUserInfo(r)

	links, beforelink := getlinks(userinfo, feedid, lastlink, search)

	templinfo := make(map[string]interface{})
	templinfo["Links"] = links
	templinfo["BeforeLink"] = beforelink
	err := readviews.ExecuteTemplate(w, "links.html", templinfo)
	if err != nil {
		log.Printf("template error: %s", err)
	}
}

func domarkseen(w http.ResponseWriter, r *http.Request) {
	linkid, _ := strconv.ParseInt(mux.Vars(r)["linkid"], 10, 0)
	userinfo := GetUserInfo(r)
	var err error
	if linkid > 0 {
		_, err = db.Exec("update userlinks set seen = 1 where userid = ? and linkid = ?",
			userinfo.UserID, linkid)
		if linkid == userinfo.LastUnread {
			_, e2 := db.Exec("update users set lastunread = (select coalesce((select min(u1.linkid) from userlinks as u1 where u1.seen = 0 and u1.userid = ?), (select max(u2.linkid) + 1 from userlinks as u2 where u2.seen = 1 and u2.userid = ?)));",
				userinfo.UserID, userinfo.UserID)
			if e2 != nil {
				log.Printf("error updating lastunread: %s", e2)
			}
		}
	} else {
		feedid, _ := strconv.ParseInt(mux.Vars(r)["feedid"], 10, 0)
		if feedid > 0 {
			_, err = db.Exec("update userlinks set seen = 1 where userid = ? and feedid = ?",
				userinfo.UserID, feedid)
		}
	}
	if err != nil {
		log.Printf("error marking seen: %s", err)
	}
	w.Write([]byte("OK\n"))
}

func dosavefeed(w http.ResponseWriter, r *http.Request) {
	feedid, _ := strconv.ParseInt(r.FormValue("feedid"), 10, 0)
	feedurl := r.FormValue("feedurl")
	feedname := r.FormValue("feedname")
	checkinterval, _ := time.ParseDuration(r.FormValue("checkinterval"))
	minduration, _ := time.ParseDuration("1h")
	if checkinterval != 0 && checkinterval < minduration {
		checkinterval = minduration
	}

	log.Println("saving", feedid, feedurl, feedname)

	var err error
	if feedid > 0 {
		_, err = db.Exec("update feeds set feedurl = ?, feedname = ?, checkinterval = ?, nextcheck = ? where feedid = ?",
			feedurl, feedname, checkinterval.String(), "", feedid)
	} else {
		_, err = db.Exec("insert into feeds (feedurl, feedname, checkinterval, nextcheck) values (?, ?, ?, ?)",
			feedurl, feedname, "12h", "")
	}
	if err != nil {
		log.Printf("error saving feed: %s", err)
	}
	go func() { wakefeeder <- 0 }()
	http.Redirect(w, r, "/feeds", http.StatusSeeOther)
}

func servecss(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Cache-Control", "max-age=432000")
	http.ServeFile(w, r, "views"+r.URL.Path)
}
func servehtml(w http.ResponseWriter, r *http.Request) {
	w.Header().Set("Cache-Control", "max-age=432000")
	http.ServeFile(w, r, "views"+r.URL.Path+".html")
}
func serveimage(w http.ResponseWriter, r *http.Request) {
	name := mux.Vars(r)["name"]
	rows, err := db.Query("select data from images where name = ?", name)
	if err != nil {
		log.Printf("error querying images: %s", err)
		return
	}
	defer rows.Close()
	if rows.Next() {
		var data sql.RawBytes
		err = rows.Scan(&data)
		if err != nil {
			log.Printf("error loading data: %s", err)
			return
		}
		w.Header().Set("Cache-Control", "max-age=432000")
		w.Write(data)
	}
}

func reader() {
	mux := mux.NewRouter()
	getters := mux.Methods("GET").Subrouter()
	getters.HandleFunc("/", LoginWrap(linkspage))
	getters.HandleFunc("/before/{linkid:[0-9]+}", LoginWrap(linkspage))
	getters.HandleFunc("/feeds", LoginWrap(feedspage))
	getters.HandleFunc("/feed/{feedid:[0-9]+}", LoginWrap(linkspage))
	getters.HandleFunc("/feed/{feedid:[0-9]+}/before/{linkid:[0-9]+}", LoginWrap(linkspage))
	getters.HandleFunc("/images/{name:[[:alnum:]]+}.{format:[[:alnum:]]+}", LoginWrap(serveimage))
	getters.HandleFunc("/search", LoginWrap(linkspage))
	getters.HandleFunc("/search/before/{linkid:[0-9]+}", LoginWrap(linkspage))
	getters.HandleFunc("/style.css", servecss)
	getters.HandleFunc("/login", servehtml)
	getters.HandleFunc("/logout", LoginWrap(dologout))
	posters := mux.Methods("POST").Subrouter()
	posters.HandleFunc("/mark/{linkid:[0-9]+}", LoginWrap(domarkseen))
	posters.HandleFunc("/markall/{feedid:[0-9]+}", LoginWrap(domarkseen))
	posters.HandleFunc("/savefeed", LoginWrap(dosavefeed))
	posters.HandleFunc("/dologin", dologin)

	serv := &http.Server{
		Addr:    "127.0.0.1:8451",
		Handler: mux,
	}
	serv.ListenAndServe()
}
